import { EmbeddedViewRef, IterableChanges, NgZone } from '@angular/core';
import { RxStrategyCredentials } from '@rx-angular/cdk/render-strategies';
import { RxListViewContext } from '@rx-angular/cdk/template';
import { Observable } from 'rxjs';
import { ListRange, ListTemplateChange, ListTemplateChangeType, TemplateSettings } from './model';
/**
 * @internal
 *
 * An object that holds methods needed to introduce actions to a list e.g. move, remove, insert
 */
export interface RxVirtualListTemplateManager<T, C> {
    toTemplateWork(iterableChanges: IterableChanges<T>, items: T[], renderedRange: ListRange, strategy: RxStrategyCredentials, count: number, ngZone?: NgZone): {
        insertedOrRemoved: boolean;
        changes: ListTemplateChange<T>[];
        work$: Observable<ListTemplateChange<T>>[];
    };
    viewRendered$: Observable<{
        view: EmbeddedViewRef<C>;
        index?: number;
        item: T;
        change: ListTemplateChangeType;
    }>;
    detach(): void;
}
/**
 * @internal
 *
 * Factory that returns a `ListTemplateManager` for the passed params.
 *
 * @param templateSettings
 */
export declare function getVirtualTemplateHandler<C extends RxListViewContext<T>, T>({ viewContainerRef, templateRef, createViewContext, updateViewContext, viewCacheSize, }: TemplateSettings<T, C, any>): RxVirtualListTemplateManager<T, C>;
