import * as i0 from '@angular/core';
import { Directive, Output, InjectionToken, Optional, Inject, Input, NgModule, Component, ChangeDetectionStrategy, ViewChild, ContentChild } from '@angular/core';
import { RxDefaultListViewContext } from '@rx-angular/cdk/template';
import { Subject, Observable, from, ReplaySubject, combineLatest, merge, of, NEVER, concat, defer } from 'rxjs';
import { coalesceWith } from '@rx-angular/cdk/coalescing';
import { distinctUntilChanged, takeUntil, filter, map, startWith, tap, switchMap, debounce, catchError, ignoreElements, shareReplay } from 'rxjs/operators';
import { requestAnimationFrame, cancelAnimationFrame, Promise as Promise$1 } from '@rx-angular/cdk/zone-less/browser';
import { coerceDistinctWith } from '@rx-angular/cdk/coercing';
import * as i2 from '@rx-angular/cdk/render-strategies';
import { onStrategy, strategyHandling } from '@rx-angular/cdk/render-strategies';

/**
 * @Directive RxVirtualScrollStrategy
 *
 * @description
 * Abstract implementation for the actual implementations of the ScrollStrategies
 * being consumed by `*rxVirtualFor` and `rx-virtual-scroll-viewport`.
 *
 * This is one of the core parts for the virtual scrolling implementation. It has
 * to determine the `ListRange` being rendered to the DOM as well as managing
 * the layouting task for the `*rxVirtualFor` directive.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
class RxVirtualScrollStrategy {
    constructor() {
        /**
         * @description
         *
         * Emits whenever an update to a single view was rendered
         */
        this.viewRenderCallback = new Subject();
    }
    /** @internal */
    getElement(view) {
        if (this.nodeIndex !== undefined) {
            return view.rootNodes[this.nodeIndex];
        }
        const rootNode = view.rootNodes[0];
        this.nodeIndex = rootNode instanceof HTMLElement ? 0 : 1;
        return view.rootNodes[this.nodeIndex];
    }
}
/** @nocollapse */ RxVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollStrategy, deps: [], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ RxVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualScrollStrategy, outputs: { viewRenderCallback: "viewRenderCallback" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollStrategy, decorators: [{
            type: Directive
        }], propDecorators: { viewRenderCallback: [{
                type: Output
            }] } });
/** @internal */
class RxVirtualScrollViewport {
}
/** @nocollapse */ RxVirtualScrollViewport.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewport, deps: [], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ RxVirtualScrollViewport.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualScrollViewport, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewport, decorators: [{
            type: Directive
        }] });
/** @internal */
class RxVirtualViewRepeater {
    constructor() {
        this._trackBy = (i, a) => a;
    }
}
/** @nocollapse */ RxVirtualViewRepeater.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualViewRepeater, deps: [], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ RxVirtualViewRepeater.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualViewRepeater, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualViewRepeater, decorators: [{
            type: Directive
        }] });
/** @internal */
class RxVirtualForViewContext extends RxDefaultListViewContext {
    constructor(item, rxVirtualForOf, customProps) {
        super(item, customProps);
        this.rxVirtualForOf = rxVirtualForOf;
    }
}

function unpatchedAnimationFrameTick() {
    return new Observable((observer) => {
        const tick = requestAnimationFrame(() => {
            observer.next();
            observer.complete();
        });
        return () => {
            cancelAnimationFrame(tick);
        };
    });
}
function unpatchedMicroTask() {
    return from(Promise$1.resolve());
}
function getZoneUnPatchedApi(targetOrName, name) {
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    return targetOrName['__zone_symbol__' + name] || targetOrName[name];
}

/** Injection token to be used to override the default options. */
const RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS = new InjectionToken('rx-virtual-scrolling-default-options', {
    providedIn: 'root',
    factory: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS_FACTORY,
});
/** @internal */
function RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS_FACTORY() {
    return {
        runwayItems: DEFAULT_RUNWAY_ITEMS,
        runwayItemsOpposite: DEFAULT_RUNWAY_ITEMS_OPPOSITE,
        viewCacheSize: DEFAULT_VIEW_CACHE_SIZE,
        itemSize: DEFAULT_ITEM_SIZE,
    };
}
/** @internal */
const DEFAULT_VIEW_CACHE_SIZE = 20;
/** @internal */
const DEFAULT_ITEM_SIZE = 50;
/** @internal */
const DEFAULT_RUNWAY_ITEMS = 10;
/** @internal */
const DEFAULT_RUNWAY_ITEMS_OPPOSITE = 2;

/** @internal */
function removeFromArray$1(arr, index) {
    // perf: array.pop is faster than array.splice!
    if (index >= arr.length - 1) {
        return arr.pop();
    }
    else {
        return arr.splice(index, 1)[0];
    }
}
const defaultSizeExtract = (entry) => entry.borderBoxSize[0].blockSize;
/**
 * @Directive AutosizeVirtualScrollStrategy
 *
 * @description
 *
 * The `AutosizeVirtualScrollStrategy` provides a twitter-like virtual-scrolling
 * experience. It is able to render and position items based on their individual
 * size. It is comparable to \@angular/cdk/experimental `AutosizeVirtualScrollStrategy`, but
 * with a high performant layouting technique and more features.
 *
 * On top of this the `AutosizeVirtualScrollStrategy` is leveraging the native
 * `ResizeObserver` in order to detect size changes for each individual view
 * rendered to the DOM and properly re-position accordingly.
 *
 * In order to provide top-notch runtime performance the `AutosizeVirtualScrollStrategy`
 * builds up caches that prevent DOM interactions whenever possible. Once a view
 * was visited, its properties will be stored instead of re-read from the DOM again as
 * this can potentially lead to unwanted forced reflows.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class AutosizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    /** @internal */
    constructor(differs, defaults) {
        var _a, _b, _c, _d, _e, _f;
        super();
        this.differs = differs;
        this.defaults = defaults;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        /**
         * @description
         * The default size of the items being rendered. The autosized strategy will assume
         * this size for items it doesn't know yet. For the smoothest experience,
         * you provide the mean size of all items being rendered - if possible of course.
         *
         * As soon as rxVirtualFor is able to also render actual tombstone items, this
         * will be the size of a tombstone item being rendered before the actual item
         * is inserted into its position.
         */
        this.tombstoneSize = (_f = (_e = this.defaults) === null || _e === void 0 ? void 0 : _e.itemSize) !== null && _f !== void 0 ? _f : DEFAULT_ITEM_SIZE;
        /** @internal */
        this.viewport = null;
        /** @internal */
        this.viewRepeater = null;
        /** @internal */
        this.dataDiffer = null;
        /** @internal */
        this._contentSize$ = new ReplaySubject(1);
        /** @internal */
        this.contentSize$ = this._contentSize$.asObservable();
        /** @internal */
        this._contentSize = 0;
        /** @internal */
        this._renderedRange$ = new ReplaySubject(1);
        /** @internal */
        this.renderedRange$ = this._renderedRange$.asObservable();
        /** @internal */
        this._renderedRange = { start: 0, end: 0 };
        // range of items where size is known and doesn't need to be re-calculated
        /** @internal */
        this._cachedSet = new Set();
        /** @internal */
        this._scrolledIndex$ = new ReplaySubject(1);
        /** @internal */
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        /** @internal */
        this._scrolledIndex = 0;
        /**
         * is set, when scrollToIndex is called and the scroll event touches
         * the edge of the viewport at the bottom. This is the only event where
         * we actually have to `guess` the scroll anchor and might end up having
         * to adjust
         * @internal
         * */
        this._scrollToEdgeIndex = null;
        /**
         * is set, when scrollToIndex is called
         * @internal
         * */
        this._scrollToAnchor = null;
        /** @internal */
        this.containerSize = 0;
        /** @internal */
        this.contentLength = 0;
        /** @internal */
        this._virtualItems = [];
        /** @internal */
        this.scrollTop = 0;
        /** @internal */
        this.direction = 'down';
        /** @internal */
        this.anchorScrollTop = 0;
        /** @internal */
        this.anchorItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.lastScreenItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.detached$ = new Subject();
        /** @internal */
        this.viewsResized$ = new Subject();
        /** @internal */
        this.afterViewResized$ = new Subject();
        /** @internal */
        this.runwayStateChanged$ = new Subject();
    }
    /** @internal */
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    /** @internal */
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    /** @internal */
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    get scrolledIndex() {
        return this._scrolledIndex;
    }
    /** @internal */
    set scrolledIndex(index) {
        this._scrolledIndex = index;
        this._scrolledIndex$.next(index);
    }
    /** @internal */
    until$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    get extractSize() {
        var _a, _b;
        return (_b = (_a = this.resizeObserverConfig) === null || _a === void 0 ? void 0 : _a.extractSize) !== null && _b !== void 0 ? _b : defaultSizeExtract;
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    /** @internal */
    ngOnDestroy() {
        this.detach();
    }
    /** @internal */
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.resizeObserver = new ResizeObserver((events) => {
            this.viewsResized$.next(events);
        });
        this.maintainVirtualItems();
        this.calcRenderedRange();
        this.positionElements();
    }
    /** @internal */
    detach() {
        var _a;
        this.viewport = null;
        this.viewRepeater = null;
        this._virtualItems = [];
        (_a = this.resizeObserver) === null || _a === void 0 ? void 0 : _a.disconnect();
        this.resizeObserver = undefined;
        this.detached$.next();
    }
    scrollToIndex(index, behavior) {
        const _index = Math.min(Math.max(index, 0), this.contentLength - 1);
        let size = 0;
        for (let i = 0; i < _index; i++) {
            size += this.getItemSize(i);
        }
        this._scrollToAnchor = {
            index: _index,
            offset: 0,
        };
        this._scrollToEdgeIndex = null;
        if (size + this.containerSize > this._contentSize) {
            this._scrollToEdgeIndex = _index;
            this._scrollToAnchor = this.calculateAnchoredItem(this._scrollToAnchor, -Math.max(size + this.containerSize - this._contentSize));
        }
        this.viewport.scrollTo(size, behavior);
    }
    /** @internal */
    maintainVirtualItems() {
        // reset virtual viewport when opposite orientation to the scroll direction
        // changes, as we have to expect dimension changes for all items when this
        // happens. This could also be configurable as it maybe costs performance
        this.viewport.containerRect$.pipe(filter(() => this._virtualItems.length > 0), map(({ width }) => width), distinctUntilChanged(), this.until$()).subscribe(() => {
            // reset because we have no idea how items will behave
            let i = 0;
            let size = 0;
            this._cachedSet.clear();
            while (i < this.renderedRange.start) {
                this._virtualItems[i].size = 0;
                i++;
                size += this.tombstoneSize;
            }
            while (i >= this.renderedRange.start && i < this.renderedRange.end) {
                size += this._virtualItems[i].size;
                this._cachedSet.add(i);
                i++;
            }
            while (i < this.contentLength - 1) {
                this._virtualItems[i].size = 0;
                i++;
                size += this.tombstoneSize;
            }
            this.contentSize = size;
        });
        this.viewRepeater.values$.pipe(this.until$()).subscribe((items) => {
            var _a;
            const changes = (_a = this.getDiffer(items)) === null || _a === void 0 ? void 0 : _a.diff(items);
            if (changes) {
                // reset cache on update
                changes.forEachOperation((item, adjustedPreviousIndex, currentIndex) => {
                    if (item.previousIndex == null) {
                        const entry = {
                            size: 0,
                        };
                        if (currentIndex !== null &&
                            currentIndex < this._virtualItems.length) {
                            this._virtualItems.splice(currentIndex, 0, entry);
                        }
                        else {
                            this._virtualItems.push(entry);
                        }
                    }
                    else if (currentIndex === null) {
                        const removeIdx = adjustedPreviousIndex === null
                            ? this._virtualItems.length - 1
                            : adjustedPreviousIndex;
                        removeFromArray$1(this._virtualItems, removeIdx);
                        this._cachedSet.delete(removeIdx);
                    }
                    else if (adjustedPreviousIndex !== null) {
                        this._virtualItems[currentIndex] =
                            this._virtualItems[adjustedPreviousIndex];
                    }
                });
                changes.forEachIdentityChange(({ currentIndex }) => currentIndex && this._cachedSet.delete(currentIndex));
            }
            if (!this._contentSize && this._virtualItems.length > 0) {
                this.contentSize = this._virtualItems.length * this.tombstoneSize;
            }
        });
    }
    /** @internal */
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged());
        const onScroll$ = this.viewport.elementScrolled$.pipe(coalesceWith(unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((scrollTop) => {
            this.direction = scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = scrollTop;
        }));
        combineLatest([
            dataLengthChanged$.pipe(tap((length) => {
                this.contentLength = length;
            })),
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.afterViewResized$.pipe(startWith(void 0)),
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerSize]) => {
            this.containerSize = containerSize;
            const range = { start: 0, end: 0 };
            const delta = this.scrollTop - this.anchorScrollTop;
            if (this._scrollToAnchor) {
                this.anchorItem = this._scrollToAnchor;
                this._scrollToAnchor = null;
            }
            else {
                // reset scrollToIndex after consequent scroll or any other update
                if (this._scrollToEdgeIndex !== null) {
                    this._scrollToEdgeIndex = null;
                }
                if (this.scrollTop == 0) {
                    this.anchorItem = { index: 0, offset: 0 };
                }
                else {
                    this.anchorItem = this.calculateAnchoredItem(this.anchorItem, delta);
                }
            }
            this.scrolledIndex = this.anchorItem.index;
            this.anchorScrollTop = this.scrollTop;
            this.lastScreenItem = this.calculateAnchoredItem(this.anchorItem, containerSize);
            if (this.direction === 'up') {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItems);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItemsOpposite);
            }
            else {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItemsOpposite);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItems);
            }
            return range;
        }))
            .pipe(distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.until$())
            .subscribe((range) => (this.renderedRange = range));
    }
    /** @internal */
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const renderedRange = this.renderedRange;
            const adjustIndexWith = renderedRange.start;
            let scrolledIndex = null;
            let position = 0;
            let renderedViews = [];
            return merge(this.viewRepeater.viewRendered$.pipe(tap(({ view, index: viewIndex, item }) => {
                const adjustedIndex = viewIndex + adjustIndexWith;
                // this most of the time causes a forced reflow per rendered view.
                // it doesn't sound good, but it's still way more stable than
                // having one large reflow in a microtask after the actual
                // scheduler tick.
                // Right here, we can insert work into the task which is currently
                // executed as part of the concurrent scheduler tick.
                // causing the forced reflow here, also makes it count for the
                // schedulers frame budget. This way we will always run with the
                // configured FPS. The only case where this is not true is when rendering 1 single view
                // already explodes the budget
                this.updateElementSize(view, adjustedIndex);
                renderedViews.push({
                    view,
                    index: adjustedIndex,
                    item,
                });
            }), 
            // coalescing the layout to the next microtask
            // -> layout happens in the same tick as the scheduler. it will
            // be executed at the end and causes a forced reflow
            // to not have a reflow, we would need to let the browser stabilize the
            // current layout before reading `offsetHeight` of the nodes.
            // while the flames get smoother, other issues arise with that approach.
            // but here is def. room for improvement
            coalesceWith(unpatchedMicroTask()), map(() => {
                if (position === 0) {
                    position = this.calcInitialPosition(renderedRange);
                }
                let lastIndex = 0;
                let scrollTo = this.anchorScrollTop;
                // update DOM happens here
                // layout should be stable by now, we are thrashing it again, though
                for (let i = 0; i < renderedViews.length; i++) {
                    const { view, index, item } = renderedViews[i];
                    const element = this.getElement(view);
                    this.positionElement(element, position);
                    if (index === this._scrollToEdgeIndex) {
                        scrollTo = position;
                        this._scrollToEdgeIndex = null;
                    }
                    position += this._virtualItems[index].size;
                    if (scrolledIndex == null && position > this.scrollTop) {
                        scrolledIndex = index;
                    }
                    lastIndex = index;
                    this.viewRenderCallback.next({
                        index,
                        view,
                        item,
                    });
                }
                renderedViews = [];
                // this condition only covers the edge case when we scroll to the
                // lower edge of the viewport and we have to wait until the whole
                // renderedRange actually got rendered
                if (this._scrollToEdgeIndex === null) {
                    this.contentSize =
                        position + this.getRemainingSizeFrom(lastIndex + 1);
                    if (scrolledIndex != null &&
                        scrolledIndex !== this.scrolledIndex) {
                        this.scrolledIndex = scrolledIndex;
                    }
                    if (scrollTo !== this.scrollTop) {
                        const maxScrollTo = this._contentSize - this.containerSize;
                        if (scrollTo >= maxScrollTo &&
                            Math.ceil(this.scrollTop) >= maxScrollTo) {
                            // just trigger re-calculation of the renderedRange in case this happens
                            this.afterViewResized$.next();
                        }
                        else {
                            this.viewport.scrollTo(scrollTo);
                        }
                    }
                    else if (position < this.containerSize &&
                        lastIndex < this.contentLength) {
                        this.afterViewResized$.next();
                    }
                }
            })), this.viewRepeater.viewsRendered$.pipe(switchMap((views) => this.observeViewSizes$(adjustIndexWith, views).pipe(tap((lowestId) => {
                let i = lowestId;
                const range = {
                    start: i + adjustIndexWith,
                    end: renderedRange.end,
                };
                let position = this.calcInitialPosition(range);
                let index = range.start;
                for (i; i < views.length; i++) {
                    const element = this.getElement(views[i]);
                    this.positionElement(element, position);
                    index = i + adjustIndexWith;
                    position += this.getItemSize(index);
                }
                this.contentSize =
                    position + this.getRemainingSizeFrom(index + 1);
                this.afterViewResized$.next();
            })))));
        }), this.until$()).subscribe();
    }
    /**
     * @internal
     * heavily inspired by
     *   https://github.com/GoogleChromeLabs/ui-element-samples/blob/gh-pages/infinite-scroller/scripts/infinite-scroll.js
     */
    calculateAnchoredItem(initialAnchor, delta) {
        var _a;
        if (delta == 0)
            return initialAnchor;
        delta += initialAnchor.offset;
        let i = initialAnchor.index;
        let tombstones = 0;
        const items = this._virtualItems;
        if (delta < 0) {
            while (delta < 0 && i > 0 && items[i - 1].size) {
                delta += items[i - 1].size;
                i--;
            }
            tombstones = Math.max(-i, Math.ceil(Math.min(delta, 0) / this.tombstoneSize));
        }
        else {
            while (delta > 0 &&
                i < this.contentLength &&
                ((_a = items[i]) === null || _a === void 0 ? void 0 : _a.size) &&
                items[i].size < delta) {
                delta -= items[i].size;
                i++;
            }
            if (i >= this.contentLength) {
                tombstones = 0;
            }
            else if (!items[i].size) {
                tombstones = Math.floor(Math.max(delta, 0) / this.tombstoneSize);
            }
        }
        i += tombstones;
        delta -= tombstones * this.tombstoneSize;
        return {
            index: Math.min(i, this.contentLength),
            offset: Math.max(0, delta),
        };
    }
    /** @internal */
    calcInitialPosition(range) {
        let pos = 0;
        let i = 0;
        this.anchorScrollTop = 0;
        for (i = 0; i < this.anchorItem.index; i++) {
            this.anchorScrollTop += this.getItemSize(i);
        }
        this.anchorScrollTop += this.anchorItem.offset;
        // Calculate position of starting node
        pos = this.anchorScrollTop - this.anchorItem.offset;
        i = this.anchorItem.index;
        while (i > range.start) {
            const itemSize = this.getItemSize(i - 1);
            pos -= itemSize;
            i--;
        }
        while (i < range.start) {
            const itemSize = this.getItemSize(i);
            pos += itemSize;
            i++;
        }
        return pos;
    }
    /** @internal */
    updateElementSize(view, index) {
        const oldSize = this._virtualItems[index].size;
        const isCached = this._cachedSet.has(index) && oldSize > 0;
        const size = isCached
            ? oldSize
            : this.getElementSize(this.getElement(view));
        this._virtualItems[index] = { size };
        this._cachedSet.add(index);
        return size;
    }
    /** @internal */
    observeViewSizes$(adjustIndexWith, views) {
        var _a;
        const elementCache = new WeakMap();
        let lowestResizedId;
        for (let i = 0; i < views.length; i++) {
            const element = this.getElement(views[i]);
            this.resizeObserver.observe(element, (_a = this.resizeObserverConfig) === null || _a === void 0 ? void 0 : _a.options);
            elementCache.set(element, i);
        }
        return new Observable((observer) => {
            const inner = this.viewsResized$.subscribe((events) => {
                events.forEach((event) => {
                    if (!event.target.isConnected) {
                        return;
                    }
                    const cachedId = elementCache.get(event.target);
                    if (cachedId !== undefined) {
                        const adjustedId = cachedId + adjustIndexWith;
                        const size = Math.round(this.extractSize(event));
                        if (this._virtualItems[adjustedId].size !== size) {
                            lowestResizedId = Math.min(lowestResizedId !== null && lowestResizedId !== void 0 ? lowestResizedId : Number.MAX_SAFE_INTEGER, cachedId);
                            this._virtualItems[adjustedId].size = size;
                        }
                    }
                });
                if (lowestResizedId !== undefined) {
                    observer.next(lowestResizedId);
                }
            });
            return () => {
                var _a;
                for (let i = 0; i < views.length; i++) {
                    const element = this.getElement(views[i]);
                    (_a = this.resizeObserver) === null || _a === void 0 ? void 0 : _a.unobserve(element);
                }
                inner.unsubscribe();
            };
        }).pipe(coalesceWith(unpatchedAnimationFrameTick()), map((lowestId) => {
            lowestResizedId = undefined;
            return lowestId;
        }));
    }
    /** @internal */
    getRemainingSizeFrom(from) {
        let remaining = 0;
        for (let i = from; i < this.contentLength; i++) {
            remaining += this.getItemSize(i);
        }
        return remaining;
    }
    /** @internal */
    getItemSize(index) {
        return this._virtualItems[index].size || this.tombstoneSize;
    }
    /** @internal */
    getElementSize(element) {
        return element.offsetHeight;
    }
    /** @internal */
    positionElement(element, scrollTop) {
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
    /** @internal */
    getDiffer(values) {
        if (this.dataDiffer) {
            return this.dataDiffer;
        }
        return values
            ? (this.dataDiffer = this.differs
                .find(values)
                .create(this.viewRepeater._trackBy))
            : null;
    }
}
/** @nocollapse */ AutosizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategy, deps: [{ token: i0.IterableDiffers }, { token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ AutosizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: AutosizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[autosize]", inputs: { runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite", tombstoneSize: "tombstoneSize", resizeObserverConfig: "resizeObserverConfig" }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: AutosizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[autosize]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: AutosizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: i0.IterableDiffers }, { type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }], tombstoneSize: [{
                type: Input
            }], resizeObserverConfig: [{
                type: Input
            }] } });
class AutosizeVirtualScrollStrategyModule {
}
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, declarations: [AutosizeVirtualScrollStrategy], exports: [AutosizeVirtualScrollStrategy] });
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [AutosizeVirtualScrollStrategy],
                    declarations: [AutosizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });

/** @internal */
function removeFromArray(arr, index) {
    // perf: array.pop is faster than array.splice!
    if (index >= arr.length - 1) {
        return arr.pop();
    }
    else {
        return arr.splice(index, 1)[0];
    }
}
const defaultItemSize = () => DEFAULT_ITEM_SIZE;
/**
 * @Directive DynamicSizeVirtualScrollStrategy
 *
 * @description
 *
 * The `DynamicSizeVirtualScrollStrategy` is very similar to the `AutosizeVirtualScrollStrategy`.
 * It positions items based on a function determining its size.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class DynamicSizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    /** @internal */
    constructor(differs, defaults) {
        var _a, _b, _c, _d;
        super();
        this.differs = differs;
        this.defaults = defaults;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        this._itemSizeFn = defaultItemSize;
        /** @internal */
        this.viewport = null;
        /** @internal */
        this.viewRepeater = null;
        /** @internal */
        this.dataDiffer = null;
        /** @internal */
        this._contentSize$ = new ReplaySubject(1);
        /** @internal */
        this.contentSize$ = this._contentSize$.asObservable();
        /** @internal */
        this._contentSize = 0;
        /** @internal */
        this._renderedRange$ = new ReplaySubject(1);
        /** @internal */
        this.renderedRange$ = this._renderedRange$.asObservable();
        /** @internal */
        this._renderedRange = { start: 0, end: 0 };
        /** @internal */
        this._scrolledIndex$ = new ReplaySubject(1);
        /** @internal */
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        /** @internal */
        this.contentLength = 0;
        /** @internal */
        this._virtualItems = [];
        /** @internal */
        this.scrollTop = 0;
        /** @internal */
        this.direction = 'down';
        /** @internal */
        this.anchorScrollTop = 0;
        /** @internal */
        this.anchorItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.lastScreenItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.detached$ = new Subject();
        /** @internal */
        this.runwayStateChanged$ = new Subject();
    }
    /**
     * @description
     * Function returning the size of an item
     */
    set itemSize(fn) {
        if (fn) {
            this._itemSizeFn = fn;
        }
    }
    get itemSize() {
        return this._itemSizeFn;
    }
    /** @internal */
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    // range of items where size is known and doesn't need to be re-calculated
    /** @internal */
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    /** @internal */
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    set scrolledIndex(index) {
        this._scrolledIndex$.next(index);
    }
    /** @internal */
    until$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    /** @internal */
    ngOnDestroy() {
        this.detach();
    }
    /** @internal */
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.maintainVirtualItems();
        this.calcRenderedRange();
        this.positionElements();
    }
    /** @internal */
    detach() {
        this.viewport = null;
        this.viewRepeater = null;
        this._virtualItems = [];
        this.detached$.next();
    }
    scrollToIndex(index, behavior) {
        const _index = Math.min(Math.max(index, 0), this.contentLength - 1);
        let offset = 0;
        for (let i = 0; i < _index; i++) {
            offset += this._virtualItems[i].size;
        }
        this.viewport.scrollTo(offset, behavior);
    }
    /** @internal */
    maintainVirtualItems() {
        this.viewRepeater.values$.pipe(this.until$()).subscribe((items) => {
            var _a;
            const changes = (_a = this.getDiffer(items)) === null || _a === void 0 ? void 0 : _a.diff(items);
            if (changes) {
                changes.forEachOperation((change, adjustedPreviousIndex, currentIndex) => {
                    if (change.previousIndex == null) {
                        const entry = {
                            size: this.itemSize(change.item),
                        };
                        if (currentIndex !== null &&
                            currentIndex < this._virtualItems.length) {
                            this._virtualItems.splice(currentIndex, 0, entry);
                        }
                        else {
                            this._virtualItems.push(entry);
                        }
                    }
                    else if (currentIndex == null) {
                        const removeIdx = adjustedPreviousIndex == null
                            ? this._virtualItems.length - 1
                            : adjustedPreviousIndex;
                        removeFromArray(this._virtualItems, removeIdx);
                    }
                    else if (adjustedPreviousIndex !== null) {
                        this._virtualItems[currentIndex] =
                            this._virtualItems[adjustedPreviousIndex];
                    }
                });
                changes.forEachIdentityChange((record) => {
                    this._virtualItems[record.currentIndex] = {
                        size: this.itemSize(record.item),
                    };
                });
                let contentSize = 0;
                for (let i = 0; i < this._virtualItems.length; i++) {
                    contentSize += this._virtualItems[i].size;
                }
                this.contentSize = contentSize;
            }
        });
    }
    /** @internal */
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged());
        const onScroll$ = this.viewport.elementScrolled$.pipe(debounce(() => unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((_scrollTop) => {
            this.direction = _scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = _scrollTop;
        }));
        combineLatest([
            dataLengthChanged$.pipe(tap((length) => {
                this.contentLength = length;
            })),
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerHeight]) => {
            const range = { start: 0, end: 0 };
            const delta = this.scrollTop - this.anchorScrollTop;
            if (this.scrollTop == 0) {
                this.anchorItem = { index: 0, offset: 0 };
            }
            else {
                this.anchorItem = this.calculateAnchoredItem(this.anchorItem, delta);
            }
            this.scrolledIndex = this.anchorItem.index;
            this.anchorScrollTop = this.scrollTop;
            this.lastScreenItem = this.calculateAnchoredItem(this.anchorItem, containerHeight);
            if (this.direction === 'up') {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItems);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItemsOpposite);
            }
            else {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItemsOpposite);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItems);
            }
            return range;
        }))
            .pipe(distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.until$())
            .subscribe((range) => (this.renderedRange = range));
    }
    /** @internal */
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const renderedRange = this.renderedRange;
            const adjustIndexWith = renderedRange.start;
            let scrolledIndex = null;
            let position = 0;
            return this.viewRepeater.viewRendered$.pipe(map(({ view, index: viewIndex, item }, idx) => {
                const index = viewIndex + adjustIndexWith;
                if (idx === 0) {
                    position = this.calcInitialPosition(renderedRange);
                }
                const size = this.getItemSize(index);
                this.positionElement(this.getElement(view), position);
                position += size;
                if (scrolledIndex == null && position > this.scrollTop) {
                    scrolledIndex = index;
                }
                if (scrolledIndex != null) {
                    this.scrolledIndex = scrolledIndex;
                }
                this.viewRenderCallback.next({
                    index,
                    view,
                    item,
                });
            }));
        }), this.until$()).subscribe();
    }
    /**
     * @internal
     * heavily inspired by
     *   https://github.com/GoogleChromeLabs/ui-element-samples/blob/gh-pages/infinite-scroller/scripts/infinite-scroll.js
     */
    calculateAnchoredItem(initialAnchor, delta) {
        if (delta == 0)
            return initialAnchor;
        delta += initialAnchor.offset;
        let i = initialAnchor.index;
        const items = this._virtualItems;
        if (delta < 0) {
            while (delta < 0 && i > 0) {
                delta += items[i - 1].size;
                i--;
            }
        }
        else {
            while (delta > 0 && i < this.contentLength && items[i].size < delta) {
                delta -= items[i].size;
                i++;
            }
        }
        return {
            index: Math.min(i, this.contentLength),
            offset: delta,
        };
    }
    /** @internal */
    calcInitialPosition(range) {
        let pos = 0;
        let i = 0;
        this.anchorScrollTop = 0;
        for (i = 0; i < this.anchorItem.index; i++) {
            this.anchorScrollTop += this.getItemSize(i);
        }
        this.anchorScrollTop += this.anchorItem.offset;
        // Calculate position of starting node
        pos = this.anchorScrollTop - this.anchorItem.offset;
        i = this.anchorItem.index;
        while (i > range.start) {
            const itemSize = this.getItemSize(i - 1);
            pos -= itemSize;
            i--;
        }
        while (i < range.start) {
            const itemSize = this.getItemSize(i);
            pos += itemSize;
            i++;
        }
        return pos;
    }
    /** @internal */
    getItemSize(index) {
        return this._virtualItems[index].size;
    }
    /** @internal */
    positionElement(element, scrollTop) {
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
    /** @internal */
    getDiffer(values) {
        if (this.dataDiffer) {
            return this.dataDiffer;
        }
        return values
            ? (this.dataDiffer = this.differs
                .find(values)
                .create(this.viewRepeater._trackBy))
            : null;
    }
}
/** @nocollapse */ DynamicSizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategy, deps: [{ token: i0.IterableDiffers }, { token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ DynamicSizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: DynamicSizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[dynamic]", inputs: { runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite", itemSize: ["dynamic", "itemSize"] }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: DynamicSizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[dynamic]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: DynamicSizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: i0.IterableDiffers }, { type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }], itemSize: [{
                type: Input,
                args: ['dynamic']
            }] } });
class DynamicSizeVirtualScrollStrategyModule {
}
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, declarations: [DynamicSizeVirtualScrollStrategy], exports: [DynamicSizeVirtualScrollStrategy] });
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [DynamicSizeVirtualScrollStrategy],
                    declarations: [DynamicSizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });

/**
 * @Directive FixedSizeVirtualScrollStrategy
 *
 * @description
 *
 * The `FixedSizeVirtualScrollStrategy` provides a very performant way of rendering
 * items of a given size. It is comparable to \@angular/cdk `FixedSizeVirtualScrollStrategy`, but
 * with a high performant layouting technique.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class FixedSizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    constructor(defaults) {
        var _a, _b, _c, _d;
        super();
        this.defaults = defaults;
        this._itemSize = DEFAULT_ITEM_SIZE;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        /** @internal */
        this.runwayStateChanged$ = new Subject();
        this.viewport = null;
        this.viewRepeater = null;
        this._scrolledIndex$ = new ReplaySubject(1);
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        this._scrolledIndex = 0;
        this._contentSize$ = new ReplaySubject(1);
        this.contentSize$ = this._contentSize$.asObservable();
        this._contentSize = 0;
        this._renderedRange$ = new ReplaySubject(1);
        this.renderedRange$ = this._renderedRange$.asObservable();
        this._renderedRange = { start: 0, end: 0 };
        this.scrollTop = 0;
        this.direction = 'down';
        this.detached$ = new Subject();
    }
    /**
     * @description
     * The size of the items in the virtually scrolled list
     */
    set itemSize(itemSize) {
        if (typeof itemSize === 'number') {
            this._itemSize = itemSize;
        }
    }
    get itemSize() {
        return this._itemSize;
    }
    set scrolledIndex(index) {
        this._scrolledIndex = index;
        this._scrolledIndex$.next(index);
    }
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    ngOnDestroy() {
        this.detach();
    }
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.calcRenderedRange();
        this.positionElements();
    }
    detach() {
        this.viewport = null;
        this.viewRepeater = null;
        this.detached$.next();
    }
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const start = this.renderedRange.start;
            return this.viewRepeater.viewRendered$.pipe(tap(({ view, index, item }) => {
                this._setViewPosition(view, (index + start) * this.itemSize);
                this.viewRenderCallback.next({
                    view,
                    item,
                    index,
                });
            }));
        }), this.untilDetached$()).subscribe();
    }
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged(), tap((dataLength) => (this.contentSize = dataLength * this.itemSize)));
        const onScroll$ = this.viewport.elementScrolled$.pipe(coalesceWith(unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((_scrollTop) => {
            this.direction = _scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = _scrollTop;
        }));
        combineLatest([
            dataLengthChanged$,
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerSize]) => {
            const range = { start: 0, end: 0 };
            if (this.direction === 'up') {
                range.start = Math.floor(Math.max(0, this.scrollTop - this.runwayItems * this.itemSize) /
                    this.itemSize);
                range.end = Math.min(length, Math.ceil((this.scrollTop +
                    containerSize +
                    this.runwayItemsOpposite * this.itemSize) /
                    this.itemSize));
            }
            else {
                range.start = Math.floor(Math.max(0, this.scrollTop - this.runwayItemsOpposite * this.itemSize) / this.itemSize);
                range.end = Math.min(length, Math.ceil((this.scrollTop +
                    containerSize +
                    this.runwayItems * this.itemSize) /
                    this.itemSize));
            }
            this.scrolledIndex = Math.floor(this.scrollTop / this.itemSize);
            return range;
        }), distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.untilDetached$())
            .subscribe((range) => (this.renderedRange = range));
    }
    scrollToIndex(index, behavior) {
        const scrollTop = this.itemSize * index;
        this.viewport.scrollTo(scrollTop, behavior);
    }
    untilDetached$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    _setViewPosition(view, scrollTop) {
        const element = this.getElement(view);
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
}
/** @nocollapse */ FixedSizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategy, deps: [{ token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ FixedSizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: FixedSizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[itemSize]", inputs: { itemSize: "itemSize", runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite" }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: FixedSizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[itemSize]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: FixedSizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { itemSize: [{
                type: Input
            }], runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }] } });
class FixedSizeVirtualScrollStrategyModule {
}
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, declarations: [FixedSizeVirtualScrollStrategy], exports: [FixedSizeVirtualScrollStrategy] });
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [FixedSizeVirtualScrollStrategy],
                    declarations: [FixedSizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });

/**
 * @internal
 *
 * Factory that returns a `ListTemplateManager` for the passed params.
 *
 * @param templateSettings
 */
function getVirtualTemplateHandler({ viewContainerRef, templateRef, createViewContext, updateViewContext, viewCacheSize, }) {
    let _viewCache = [];
    const _viewRendered$ = new Subject();
    return {
        toTemplateWork,
        viewRendered$: _viewRendered$,
        detach: () => {
            for (const view of _viewCache) {
                view.destroy();
            }
            _viewCache = [];
        },
    };
    function toTemplateWork(iterableChanges, items, renderedRange, strategy, count, ngZone) {
        const listChanges = getListChanges(iterableChanges, items);
        const changes = listChanges[0].sort(([, payloadA], [, payloadB]) => {
            return (payloadA[1] || 0) - (payloadB[1] || 0);
        });
        const insertedOrRemoved = listChanges[1];
        return {
            insertedOrRemoved,
            changes,
            work$: changes.length > 0
                ? changes.map((change) => {
                    return onStrategy(change, strategy, (_change) => {
                        const [type, [item, index, adjustedPreviousIndex]] = _change;
                        const update = () => ({
                            view: viewContainerRef.get(index || 0),
                            index,
                            change: type,
                            item,
                        });
                        switch (type) {
                            case 0 /* insert */:
                                _insertView(item, index, count, renderedRange.start + index);
                                // console.log('perform insert', update());
                                _viewRendered$.next(update());
                                break;
                            case 2 /* move */:
                                _moveView(item, adjustedPreviousIndex, index, count, renderedRange.start + index);
                                // console.log('perform move', update());
                                _viewRendered$.next(update());
                                break;
                            case 1 /* remove */:
                                if (index != null) {
                                    _detachAndCacheView(index);
                                }
                                // console.log('perform remove', update());
                                // _viewRendered$.next(update());
                                break;
                            case 3 /* update */:
                                _updateView(item, index, count, renderedRange.start + index);
                                // console.log('perform update', update());
                                _viewRendered$.next(update());
                                break;
                            case 4 /* context */:
                                _updateView(item, index, count, renderedRange.start + index);
                                // console.log('perform context', update());
                                _viewRendered$.next(update());
                                break;
                        }
                    }, { ngZone });
                })
                : [],
        };
    }
    function _updateUnchangedContext(index, count, contextIndex) {
        const view = viewContainerRef.get(index);
        view.context.updateContext({
            count,
            index: contextIndex,
        });
        view.detectChanges();
    }
    function _updateView(item, index, count, contextIndex) {
        const view = viewContainerRef.get(index);
        updateViewContext(item, view, {
            count,
            index: contextIndex,
        });
        view.detectChanges();
    }
    /**
     * Inserts a view for a new item, either from the cache or by creating a new
     * one. Returns `undefined` if the item was inserted into a cached view.
     */
    function _insertView(value, currentIndex, count, contextIndex) {
        const cachedView = _insertViewFromCache(currentIndex);
        if (cachedView) {
            updateViewContext(value, cachedView, {
                count,
                index: contextIndex,
            });
            cachedView.detectChanges();
            return undefined;
        }
        const context = createViewContext(value, {
            count,
            index: contextIndex,
        });
        const view = viewContainerRef.createEmbeddedView(templateRef, context, currentIndex);
        view.detectChanges();
        return view;
    }
    /** Detaches the view at the given index and inserts into the view cache. */
    function _detachAndCacheView(index) {
        const detachedView = viewContainerRef.detach(index);
        _maybeCacheView(detachedView);
        detachedView.detectChanges();
    }
    /** Moves view at the previous index to the current index. */
    function _moveView(value, adjustedPreviousIndex, currentIndex, count, contextIndex) {
        const view = viewContainerRef.get(adjustedPreviousIndex);
        viewContainerRef.move(view, currentIndex);
        updateViewContext(value, view, {
            count,
            index: contextIndex,
        });
        view.detectChanges();
        return view;
    }
    /**
     * Cache the given detached view. If the cache is full, the view will be
     * destroyed.
     */
    function _maybeCacheView(view) {
        if (_viewCache.length < viewCacheSize) {
            _viewCache.push(view);
            return true;
        }
        else {
            const index = viewContainerRef.indexOf(view);
            // The host component could remove views from the container outside of
            // the view repeater. It's unlikely this will occur, but just in case,
            // destroy the view on its own, otherwise destroy it through the
            // container to ensure that all the references are removed.
            if (index === -1) {
                view.destroy();
            }
            else {
                viewContainerRef.remove(index);
            }
            return false;
        }
    }
    /** Inserts a recycled view from the cache at the given index. */
    function _insertViewFromCache(index) {
        const cachedView = _viewCache.pop();
        if (cachedView) {
            viewContainerRef.insert(cachedView, index);
        }
        return cachedView || null;
    }
}
/**
 * @internal
 *
 * @param changes
 * @param items
 */
function getListChanges(changes, items) {
    const changedIdxs = new Set();
    const changesArr = [];
    let notifyParent = false;
    changes.forEachOperation(({ item, previousIndex }, adjustedPreviousIndex, currentIndex) => {
        if (previousIndex == null) {
            // insert
            changesArr.push(getInsertChange(item, currentIndex));
            changedIdxs.add(item);
            notifyParent = true;
        }
        else if (currentIndex == null) {
            // remove
            changesArr.push(getRemoveChange(item, adjustedPreviousIndex));
            notifyParent = true;
        }
        else if (adjustedPreviousIndex !== null) {
            // move
            changesArr.push(getMoveChange(item, currentIndex, adjustedPreviousIndex));
            changedIdxs.add(item);
            notifyParent = true;
        }
    });
    changes.forEachIdentityChange(({ item, currentIndex }) => {
        if (currentIndex != null && !changedIdxs.has(item)) {
            changesArr.push(getUpdateChange(item, currentIndex));
            changedIdxs.add(item);
        }
    });
    if (notifyParent) {
        for (let i = 0; i < items.length; i++) {
            const item = items[i];
            if (!changedIdxs.has(item)) {
                changesArr.push(getUnchangedChange(item, i));
            }
        }
    }
    return [changesArr, notifyParent];
    // ==========
    function getMoveChange(item, currentIndex, adjustedPreviousIndex) {
        return [
            2 /* move */,
            [item, currentIndex, adjustedPreviousIndex],
        ];
    }
    function getUpdateChange(item, currentIndex) {
        return [3 /* update */, [item, currentIndex]];
    }
    function getUnchangedChange(item, index) {
        return [4 /* context */, [item, index]];
    }
    function getInsertChange(item, currentIndex) {
        return [
            0 /* insert */,
            [item, currentIndex === null ? undefined : currentIndex],
        ];
    }
    function getRemoveChange(item, adjustedPreviousIndex) {
        return [
            1 /* remove */,
            [
                item,
                adjustedPreviousIndex === null ? undefined : adjustedPreviousIndex,
            ],
        ];
    }
}

function createVirtualListManager(config) {
    const { templateSettings, renderSettings, trackBy, iterableDiffers } = config;
    const { defaultStrategyName, strategies, cdRef: injectingViewCdRef, parent, patchZone, } = renderSettings;
    const errorHandler = renderSettings.errorHandler;
    const strategyHandling$ = strategyHandling(defaultStrategyName, strategies);
    let _differ;
    function getDiffer(values) {
        if (_differ) {
            return _differ;
        }
        return values
            ? (_differ = iterableDiffers.find(values).create(trackBy))
            : null;
    }
    //               type,  payload
    const listViewHandler = getVirtualTemplateHandler(Object.assign({}, templateSettings));
    const viewContainerRef = templateSettings.viewContainerRef;
    let partiallyFinished = false;
    let renderedRange;
    const _viewsRendered$ = new Subject();
    const _viewRendered$ = new Subject();
    const _renderingStart$ = new Subject();
    const viewRenderedSub = listViewHandler.viewRendered$
        .pipe(filter(({ view, change }) => !!view && change !== 1 /* remove */), map(({ view, item, index }) => ({
        view,
        item,
        index,
    })))
        .subscribe(_viewRendered$);
    function handleError() {
        return (o$) => o$.pipe(catchError((err) => {
            partiallyFinished = false;
            errorHandler === null || errorHandler === void 0 ? void 0 : errorHandler.handleError(err);
            return of(null);
        }));
    }
    return {
        viewsRendered$: _viewsRendered$,
        viewRendered$: _viewRendered$,
        renderingStart$: _renderingStart$,
        nextStrategy(nextConfig) {
            strategyHandling$.next(nextConfig);
        },
        render(values$, range$) {
            return combineLatest([
                values$.pipe(map((values) => Array.isArray(values)
                    ? values
                    : values != null
                        ? Array.from(values)
                        : [])),
                range$,
                strategyHandling$.strategy$.pipe(distinctUntilChanged()),
            ]).pipe(render(), handleError());
        },
        detach() {
            viewRenderedSub.unsubscribe();
            listViewHandler.detach();
        },
    };
    function render() {
        return (o$) => o$.pipe(
        // map iterable to latest diff
        map(([items, range, strategy]) => {
            renderedRange = range;
            const iterable = items.slice(range.start, range.end);
            const differ = getDiffer(iterable);
            let changes = null;
            if (differ) {
                if (partiallyFinished) {
                    const currentIterable = [];
                    for (let i = 0, ilen = viewContainerRef.length; i < ilen; i++) {
                        const viewRef = viewContainerRef.get(i);
                        currentIterable[i] = viewRef.context.$implicit;
                    }
                    differ.diff(currentIterable);
                    // partiallyFinished = false;
                }
                changes = differ.diff(iterable);
            }
            return {
                changes,
                itemsToRender: iterable,
                count: items.length,
                strategy,
            };
        }), 
        // Cancel old renders
        switchMap(({ changes, count, itemsToRender, strategy }) => {
            if (!changes) {
                return NEVER;
            }
            const { work$, insertedOrRemoved } = listViewHandler.toTemplateWork(changes, itemsToRender, renderedRange, strategy, count, patchZone);
            partiallyFinished = true;
            const notifyParent = insertedOrRemoved && parent;
            _renderingStart$.next();
            return combineLatest(
            // emit after all changes are rendered
            work$.length > 0 ? work$ : [of(itemsToRender)]).pipe(tap(() => {
                partiallyFinished = false;
                const viewsRendered = [];
                const end = viewContainerRef.length;
                let i = 0;
                for (i; i < end; i++) {
                    viewsRendered.push(viewContainerRef.get(i));
                }
                _viewsRendered$.next(viewsRendered);
            }), notifyParent
                ? switchMap((v) => concat(of(v), onStrategy(injectingViewCdRef, strategy, (_v, work, options) => {
                    work(injectingViewCdRef, options.scope);
                }, {
                    ngZone: patchZone,
                    scope: injectingViewCdRef.context ||
                        injectingViewCdRef,
                }).pipe(ignoreElements())))
                : (o$) => o$, handleError(), map(() => itemsToRender));
        }));
    }
}

/**
 * @Directive RxVirtualFor
 *
 * @description
 *
 * The `*rxVirtualFor` structural directive provides a convenient and performant
 * way for rendering huge lists of items. It brings all the benefits `rxFor` does,
 * and implements virtual rendering.
 *
 * Instead of rendering every item provided, rxVirtualFor only renders what is
 * currently visible to the user, thus providing excellent runtime performance
 * for huge sets of data.
 *
 * The technique to render items is comparable to the on used by twitter and
 * explained in very much detail by @DasSurma in his blog post about the [complexities
 * of infinite scrollers](https://developer.chrome.com/blog/infinite-scroller/).
 *
 * "Each recycling of a DOM element would normally relayout the entire runway which
 * would bring us well below our target of 60 frames per second.
 * To avoid this, we are taking the burden of layout onto ourselves and use
 * absolutely positioned elements with transforms." (@DasSurma)
 *
 * ## API
 * The API is a combination of \@rx-angular/template/for &
 *  \@angular/cdk `*cdkVirtualFor`.
 * * trackBy: `(index: number, item: T) => any` | `keyof T`
 * * strategy: `string` | `Observable<string>`
 * * parent: `boolean`;
 * * renderCallback: `Subject<T[]>`
 * * viewCache: `number`
 * * (Injected) scrollStrategy: `RxVirtualScrollStrategy<T, U>`
 * * provides itself as RxVirtualViewRepeater for RxVirtualViewPortComponent to operate
 *
 * ## Features
 * * Push based architecture
 * * Comprehensive set of context variables
 * * Opt-out of `NgZone` with `patchZone`
 * * Notify when rendering of child templates is finished (`renderCallback`)
 * * Super efficient layouting with css transformations
 * * Define a viewCache in order to re-use views instead of re-creating them
 * * Configurable RxVirtualScrollStrategy<T, U> providing the core logic to calculate the viewRange and position DOM
 * Nodes
 *
 * ### Context Variables
 *
 * The following context variables are available for each template:
 *
 * - $implicit: `T` // the default variable accessed by `let val`
 * - item$: `Observable<T>` // the same value as $implicit, but as `Observable`
 * - index: `number` // current index of the item
 * - count: `number` // count of all items in the list
 * - first: `boolean` // true if the item is the first in the list
 * - last: `boolean` // true if the item is the last in the list
 * - even: `boolean` // true if the item has on even index (index % 2 === 0)
 * - odd: `boolean` // the opposite of even
 * - index$: `Observable<number>` // index as `Observable`
 * - count$: `Observable<number>` // count as `Observable`
 * - first$: `Observable<boolean>` // first as `Observable`
 * - last$: `Observable<boolean>` // last as `Observable`
 * - even$: `Observable<boolean>` // even as `Observable`
 * - odd$: `Observable<boolean>` // odd as `Observable`
 * - select: `(keys: (keyof T)[], distinctByMap) => Observable<Partial<T>>`
 * // returns a selection function which
 * // accepts an array of properties to pluck out of every list item. The function returns the selected properties of
 * // the current list item as distinct `Observable` key-value-pair. See the example below:
 *
 * This example showcases the `select` view-context function used for deeply nested lists.
 *
 *  ```html
 * <rx-virtual-scroll-viewport>
 *   <div
 *    autosized
 *    *rxVirtualFor="let hero of heroes$; trackBy: trackItem; let select = select;">
 *     <div>
 *       <strong>{{ hero.name }}</strong></br>
 *       Defeated enemies:
 *     </div>
 *      <span *rxFor="let enemy of select(['defeatedEnemies']); trackBy: trackEnemy;">
 *        {{ enemy.name }}
 *      </span>
 *   </div>
 * </rx-virtual-scroll-viewport>
 *  ```
 *
 * ### Using the context variables
 *
 * ```html
 * <rx-virtual-scroll-viewport>
 *  <div
 *     *rxVirtualFor="
 *       let item of observableItems$;
 *       let count = count;
 *       let index = index;
 *       let first = first;
 *       let last = last;
 *       let even = even;
 *       let odd = odd;
 *       trackBy: trackItem;
 *     "
 *   >
 *     <div>{{ count }}</div>
 *     <div>{{ index }}</div>
 *     <div>{{ item }}</div>
 *     <div>{{ first }}</div>
 *     <div>{{ last }}</div>
 *     <div>{{ even }}</div>
 *     <div>{{ odd }}</div>
 *   </div>
 * </rx-virtual-scroll-viewport>
 * ```
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class RxVirtualFor {
    /** @internal */
    constructor(scrollStrategy, iterableDiffers, cdRef, ngZone, eRef, templateRef, viewContainerRef, strategyProvider, errorHandler, defaults) {
        var _a;
        this.scrollStrategy = scrollStrategy;
        this.iterableDiffers = iterableDiffers;
        this.cdRef = cdRef;
        this.ngZone = ngZone;
        this.eRef = eRef;
        this.templateRef = templateRef;
        this.viewContainerRef = viewContainerRef;
        this.strategyProvider = strategyProvider;
        this.errorHandler = errorHandler;
        this.defaults = defaults;
        /**
         * @description
         * Controls the amount if views held in cache for later re-use when a user is
         * scrolling the list. If this is set to 0, `rxVirtualFor` won't cache any view,
         * thus destroying & re-creating very often on scroll events.
         */
        this.viewCacheSize = ((_a = this.defaults) === null || _a === void 0 ? void 0 : _a.viewCacheSize) || DEFAULT_VIEW_CACHE_SIZE;
        /**
         * @description
         *  If `parent` is set to `true` (default to `false`), `*rxVirtualFor` will
         *  automatically run change-detection for its parent component when its scheduled
         *  tasks are done in order to update any pending `@ContentChild` or `@ViewChild`
         *  relation to be updated according to the updated ViewContainer.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *   <rx-virtual-scroll-viewport>
         *      <app-list-item
         *        *rxVirtualFor="
         *          let item of items$;
         *          trackBy: trackItem;
         *          parent: true;
         *        "
         *        [item]="item"
         *        autosized
         *      ></app-list-item>
         *    </rx-virtual-scroll-viewport>
         *   `
         * })
         * export class AppComponent {
         *   // those queries won't be in sync with what `rxVirtualFor` is rendering
         *   // when parent is set to false.
         *   \@ViewChildren(AppListItem) listItems: QueryList<AppListItem>;
         *
         *   items$ = itemService.getItems();
         * }
         *
         * @param renderParent
         */
        this.renderParent = false;
        /**
         * @description
         * A flag to control whether `*rxVirtualFor` rendering happens within
         * `NgZone` or not. The default value is set to `true` if not configured otherwise.
         * If `patchZone` is set to `false` `*rxVirtualFor` will operate completely outside of `NgZone`.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *    <rx-virtual-scroll-viewport>
         *      <app-list-item
         *        *rxVirtualFor="
         *          let item of items$;
         *          trackBy: trackItem;
         *          patchZone: false;
         *        "
         *        [item]="item"
         *        autosized
         *      ></app-list-item>
         *    </rx-virtual-scroll-viewport>
         *   `
         * })
         * export class AppComponent {
         *   items$ = itemService.getItems();
         * }
         *
         * @param patchZone
         */
        this.patchZone = this.strategyProvider.config.patchZone;
        /** @internal */
        this.rendered$ = new Subject();
        /** @internal */
        this.viewsRendered$ = new Subject();
        /** @internal */
        this.viewRendered$ = new Subject();
        /** @internal */
        this.renderingStart$ = new Subject();
        /** @internal */
        this.strategyInput$ = new ReplaySubject(1);
        /** @internal */
        this.observables$ = new ReplaySubject(1);
        /** @internal */
        this.values$ = this.observables$.pipe(coerceDistinctWith(), shareReplay({ bufferSize: 1, refCount: true }));
        /** @internal */
        this.strategy$ = this.strategyInput$.pipe(coerceDistinctWith());
        /** @internal */
        this._destroy$ = new Subject();
        /** @internal */
        this._trackBy = (i, a) => a;
    }
    /**
     * @description
     * The iterable input
     *
     * @example
     * <rx-virtual-scroll-viewport>
     *   <app-hero *rxVirtualFor="heroes$; let hero"
     *     [hero]="hero"></app-hero>
     * </rx-virtual-scroll-viewport>
     *
     * @param potentialObservable
     */
    set rxVirtualFor(potentialObservable) {
        this.observables$.next(potentialObservable);
    }
    /**
     * @description
     * The iterable input
     *
     * @example
     * <rx-virtual-scroll-viewport>
     *   <app-hero *rxVirtualFor="heroes$; let hero"
     *     [hero]="hero"></app-hero>
     * </rx-virtual-scroll-viewport>
     *
     * @param potentialObservable
     */
    set rxVirtualForOf(potentialObservable) {
        this.observables$.next(potentialObservable);
    }
    set rxVirtualForTemplate(value) {
        this._template = value;
    }
    /**
     * @description
     * The rendering strategy to be used to render updates to the DOM.
     * Use it to dynamically manage your rendering strategy. You can switch the strategy
     * imperatively (with a string) or by binding an Observable.
     * The default strategy is `'normal'` if not configured otherwise.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <rx-virtual-scroll-viewport>
     *       <app-hero
     *        autosized
     *        *rxVirtualFor="let hero of heroes$; strategy: strategy"
     *        [hero]="hero"></app-hero>
     *     </rx-virtual-scroll-viewport>
     *   `
     * })
     * export class AppComponent {
     *   strategy = 'low';
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <rx-virtual-scroll-viewport>
     *       <app-hero
     *        autosized
     *        *rxVirtualFor="let hero of heroes$; strategy: strategy$"
     *        [hero]="hero"></app-hero>
     *     </rx-virtual-scroll-viewport>
     *   `
     * })
     * export class AppComponent {
     *   strategy$ = new BehaviorSubject('immediate');
     * }
     *
     * @param strategyName
     * @see {@link strategies}
     */
    set strategy(strategyName) {
        this.strategyInput$.next(strategyName);
    }
    /*@Input('rxVirtualForTombstone') tombstone: TemplateRef<
     RxVirtualForViewContext<T>
     > | null = null;*/
    /**
     * @description
     * A function or key that defines how to track changes for items in the provided
     * iterable data.
     *
     * When items are added, moved, or removed in the iterable,
     * the directive must re-render the appropriate DOM nodes.
     * To minimize operations on the DOM, only nodes that have changed
     * are re-rendered.
     *
     * By default, `rxVirtualFor` assumes that the object instance identifies
     * the node in the iterable (equality check `===`).
     * When a function or key is supplied, `rxVirtualFor` uses the result to identify the item node.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <rx-virtual-scroll-viewport>
     *      <app-list-item
     *        *rxVirtualFor="
     *          let item of items$;
     *          trackBy: 'id';
     *        "
     *        autosized
     *        [item]="item"
     *      >
     *      </app-list-item>
     *    </rx-virtual-scroll-viewport>
     *   `
     * })
     * export class AppComponent {
     *   items$ = itemService.getItems();
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *   <rx-virtual-scroll-viewport>
     *      <app-list-item
     *        *rxVirtualFor="
     *          let item of items$;
     *          trackBy: trackItem;
     *        "
     *        autosized
     *        [item]="item"
     *      >
     *      </app-list-item>
     *    </rx-virtual-scroll-viewport>
     *   `
     * })
     * export class AppComponent {
     *   items$ = itemService.getItems();
     *   trackItem = (idx, item) => item.id;
     * }
     *
     * @param trackByFnOrKey
     */
    set trackBy(trackByFnOrKey) {
        if (typeof trackByFnOrKey === 'function') {
            this._trackBy = trackByFnOrKey;
        }
        else if (typeof trackByFnOrKey === 'string') {
            this._trackBy = (i, a) => a[trackByFnOrKey];
        }
        else if (ngDevMode) {
            throw new Error(`Received incorrect value for trackBy, expected string | ((idx: number, i: T) => unknown), got ${typeof trackByFnOrKey}`);
        }
    }
    /**
     * @description
     * A `Subject` which emits whenever `*rxVirtualFor` finished rendering a
     * set of changes to the view.
     * This enables developers to perform actions exactly at the timing when the
     * updates passed are rendered to the DOM.
     * The `renderCallback` is useful in situations where you rely on specific DOM
     * properties like the `height` of a table after all items got rendered.
     * It is also possible to use the renderCallback in order to determine if a
     * view should be visible or not. This way developers can hide a list as
     * long as it has not finished rendering.
     *
     * The result of the `renderCallback` will contain the currently rendered set
     * of items in the iterable.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <rx-virtual-scroll-viewport>
     *      <app-list-item
     *        *rxVirtualFor="
     *          let item of items$;
     *          trackBy: trackItem;
     *          renderCallback: itemsRendered;
     *        "
     *        autosized
     *        [item]="item"
     *      >
     *      </app-list-item>
     *    </rx-virtual-scroll-viewport>
     *   `
     * })
     * export class AppComponent {
     *   items$: Observable<Item[]> = itemService.getItems();
     *   trackItem = (idx, item) => item.id;
     *   // this emits whenever rxVirtualFor finished rendering changes
     *   itemsRendered = new Subject<Item[]>();
     * }
     *
     * @param renderCallback
     */
    set renderCallback(renderCallback) {
        this._renderCallback = renderCallback;
    }
    /** @internal */
    get template() {
        return this._template || this.templateRef;
    }
    /** @internal */
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
    /** @internal */
    ngOnInit() {
        this.listManager = createVirtualListManager({
            iterableDiffers: this.iterableDiffers,
            renderSettings: {
                cdRef: this.cdRef,
                strategies: this.strategyProvider.strategies,
                defaultStrategyName: this.strategyProvider.primaryStrategy,
                parent: this.renderParent,
                patchZone: this.patchZone ? this.ngZone : undefined,
                errorHandler: this.errorHandler,
            },
            templateSettings: {
                viewContainerRef: this.viewContainerRef,
                templateRef: this.template,
                createViewContext: this.createViewContext.bind(this),
                updateViewContext: this.updateViewContext.bind(this),
                viewCacheSize: this.viewCacheSize,
            },
            trackBy: this._trackBy,
        });
        this.listManager.nextStrategy(this.strategy$);
        this.values$.pipe(takeUntil(this._destroy$)).subscribe((values) => {
            this.values = values;
        });
        this.listManager
            .render(this.values$, this.scrollStrategy.renderedRange$)
            .pipe(takeUntil(this._destroy$))
            .subscribe((v) => {
            var _a;
            this.rendered$.next(v);
            (_a = this._renderCallback) === null || _a === void 0 ? void 0 : _a.next(v);
        });
        this.listManager.viewsRendered$
            .pipe(takeUntil(this._destroy$))
            .subscribe(this.viewsRendered$);
        this.listManager.viewRendered$
            .pipe(takeUntil(this._destroy$))
            .subscribe(this.viewRendered$);
        this.listManager.renderingStart$
            .pipe(takeUntil(this._destroy$))
            .subscribe(this.renderingStart$);
    }
    /** @internal */
    createViewContext(item, computedContext) {
        return new RxVirtualForViewContext(item, this.values, computedContext);
    }
    /** @internal */
    updateViewContext(item, view, computedContext) {
        view.context.updateContext(computedContext);
        view.context.$implicit = item;
        view.context.rxVirtualForOf = this.values;
    }
    /** @internal */
    ngOnDestroy() {
        this._destroy$.next();
        this.listManager.detach();
    }
}
/** @nocollapse */ RxVirtualFor.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualFor, deps: [{ token: RxVirtualScrollStrategy }, { token: i0.IterableDiffers }, { token: i0.ChangeDetectorRef }, { token: i0.NgZone }, { token: i0.ElementRef }, { token: i0.TemplateRef }, { token: i0.ViewContainerRef }, { token: i2.RxStrategyProvider }, { token: i0.ErrorHandler }, { token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ RxVirtualFor.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualFor, selector: "[rxVirtualFor]", inputs: { rxVirtualFor: "rxVirtualFor", rxVirtualForOf: "rxVirtualForOf", rxVirtualForTemplate: "rxVirtualForTemplate", strategy: ["rxVirtualForStrategy", "strategy"], viewCacheSize: ["rxVirtualForViewCacheSize", "viewCacheSize"], renderParent: ["rxVirtualForParent", "renderParent"], patchZone: ["rxVirtualForPatchZone", "patchZone"], trackBy: ["rxVirtualForTrackBy", "trackBy"], renderCallback: ["rxVirtualForRenderCallback", "renderCallback"] }, providers: [{ provide: RxVirtualViewRepeater, useExisting: RxVirtualFor }], ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualFor, decorators: [{
            type: Directive,
            args: [{
                    selector: '[rxVirtualFor]',
                    providers: [{ provide: RxVirtualViewRepeater, useExisting: RxVirtualFor }],
                }]
        }], ctorParameters: function () { return [{ type: RxVirtualScrollStrategy }, { type: i0.IterableDiffers }, { type: i0.ChangeDetectorRef }, { type: i0.NgZone }, { type: i0.ElementRef }, { type: i0.TemplateRef }, { type: i0.ViewContainerRef }, { type: i2.RxStrategyProvider }, { type: i0.ErrorHandler }, { type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { rxVirtualFor: [{
                type: Input
            }], rxVirtualForOf: [{
                type: Input
            }], rxVirtualForTemplate: [{
                type: Input
            }], strategy: [{
                type: Input,
                args: ['rxVirtualForStrategy']
            }], viewCacheSize: [{
                type: Input,
                args: ['rxVirtualForViewCacheSize']
            }], renderParent: [{
                type: Input,
                args: ['rxVirtualForParent']
            }], patchZone: [{
                type: Input,
                args: ['rxVirtualForPatchZone']
            }], trackBy: [{
                type: Input,
                args: ['rxVirtualForTrackBy']
            }], renderCallback: [{
                type: Input,
                args: ['rxVirtualForRenderCallback']
            }] } });

function observeElementSize(element, config) {
    var _a;
    const extractProp = (_a = config === null || config === void 0 ? void 0 : config.extract) !== null && _a !== void 0 ? _a : ((entries) => entries[0].contentRect);
    return new Observable((subscriber) => {
        const observer = new ResizeObserver((entries) => {
            subscriber.next(extractProp(entries));
        });
        observer.observe(element, config === null || config === void 0 ? void 0 : config.options);
        return () => observer.disconnect();
    });
}

/**
 * @Component RxVirtualScrollViewport
 *
 * @description
 * Container component comparable to CdkVirtualScrollViewport acting as viewport
 * for `*rxVirtualFor` to operate on.
 *
 * Its main purpose is to implement the `RxVirtualScrollViewport` interface
 * as well as maintaining the scroll runways' height in order to give
 * the provided `RxVirtualScrollStrategy` room to position items.
 *
 * Furthermore, it will gather and forward all events to the consumer of `rxVirtualFor`.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
class RxVirtualScrollViewportComponent {
    /** @internal */
    constructor(elementRef, scrollStrategy) {
        this.elementRef = elementRef;
        this.scrollStrategy = scrollStrategy;
        /** @internal */
        this.rendered$ = defer(() => this.viewRepeater.rendered$);
        /** @internal */
        this._elementScrolled = new Subject();
        this.elementScrolled$ = this._elementScrolled.asObservable();
        /** @internal */
        this._containerRect$ = new ReplaySubject(1);
        this.containerRect$ = this._containerRect$.asObservable();
        /**
         * @description
         *
         * The range to be rendered by `*rxVirtualFor`. This value is determined by the
         * provided `RxVirtualScrollStrategy`. It gives the user information about the
         * range of items being actually rendered to the DOM.
         * Note this value updates before the `renderCallback` kicks in, thus it is only
         * in sync with the DOM when the next `renderCallback` emitted an event.
         */
        this.viewRange = this.scrollStrategy.renderedRange$;
        /**
         * @description
         *
         * The index of the currently scrolled item. The scrolled item is the topmost
         * item actually being visible to the user.
         */
        this.scrolledIndexChange = this.scrollStrategy.scrolledIndex$;
        /** @internal */
        this.destroy$ = new Subject();
        this.scrollListener = (event) => this._elementScrolled.next(event);
        if (ngDevMode && !scrollStrategy) {
            throw Error('Error: rx-virtual-scroll-viewport requires an `RxVirtualScrollStrategy` to be set.');
        }
    }
    /** @internal */
    ngOnInit() {
        observeElementSize(this.elementRef.nativeElement, {
            extract: (entries) => ({
                height: Math.round(entries[0].contentRect.height),
                width: Math.round(entries[0].contentRect.width),
            }),
        })
            .pipe(distinctUntilChanged(({ height: prevHeight, width: prevWidth }, { height, width }) => prevHeight === height && prevWidth === width), takeUntil(this.destroy$))
            .subscribe(this._containerRect$);
    }
    ngAfterViewInit() {
        getZoneUnPatchedApi(this.scrollContainer(), 'addEventListener').call(this.scrollContainer(), 'scroll', this.scrollListener, {
            passive: true,
        });
    }
    /** @internal */
    ngAfterContentInit() {
        if (ngDevMode && !this.viewRepeater) {
            throw Error('Error: rx-virtual-scroll-viewport requires a `RxVirtualViewRepeater` to be provided.');
        }
        this.scrollStrategy.attach(this, this.viewRepeater);
        this.scrollStrategy.contentSize$
            .pipe(distinctUntilChanged(), takeUntil(this.destroy$))
            .subscribe((size) => this.updateContentSize(size));
    }
    /** @internal */
    ngOnDestroy() {
        getZoneUnPatchedApi(this.scrollContainer(), 'removeEventListener').call(this.scrollContainer(), 'scroll', this.scrollListener);
        this.destroy$.next();
        this.scrollStrategy.detach();
    }
    elementScrolled() {
        return this._elementScrolled.asObservable();
    }
    scrollContainer() {
        return this.scrollViewport.nativeElement;
    }
    getScrollTop() {
        return this.scrollContainer().scrollTop;
    }
    scrollTo(index, behavior) {
        // TODO: implement more complex scroll scenarios
        this.scrollContainer().scrollTo({ top: index, behavior: behavior });
    }
    scrollToIndex(index, behavior) {
        this.scrollStrategy.scrollToIndex(index, behavior);
    }
    updateContentSize(size) {
        this.runway.nativeElement.style.transform = `translate(0, ${size}px)`;
    }
}
/** @nocollapse */ RxVirtualScrollViewportComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewportComponent, deps: [{ token: i0.ElementRef }, { token: RxVirtualScrollStrategy, optional: true }], target: i0.ɵɵFactoryTarget.Component });
/** @nocollapse */ RxVirtualScrollViewportComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualScrollViewportComponent, selector: "rx-virtual-scroll-viewport", outputs: { viewRange: "viewRange", scrolledIndexChange: "scrolledIndexChange" }, providers: [
        {
            provide: RxVirtualScrollViewport,
            useExisting: RxVirtualScrollViewportComponent,
        },
    ], queries: [{ propertyName: "viewRepeater", first: true, predicate: RxVirtualViewRepeater, descendants: true }], viewQueries: [{ propertyName: "runway", first: true, predicate: ["runway"], descendants: true, static: true }, { propertyName: "scrollViewport", first: true, predicate: ["scrollViewport"], descendants: true, static: true }], ngImport: i0, template: `
    <div #scrollViewport class="rx-virtual-scroll__viewport">
      <div #runway class="rx-virtual-scroll__run-way"></div>
      <ng-content></ng-content>
    </div>
  `, isInline: true, styles: ["\n      :host {\n        display: block;\n        width: 100%;\n        height: 100%;\n        box-sizing: border-box;\n        contain: strict;\n      }\n\n      :host:not([autosize]) .rx-virtual-scroll__viewport {\n        transform: translateZ(0);\n        will-change: scroll-position;\n      }\n\n      .rx-virtual-scroll__viewport {\n        contain: strict;\n        -webkit-overflow-scrolling: touch;\n        width: 100%;\n        position: absolute;\n        top: 0;\n        bottom: 0;\n        overflow: auto;\n      }\n\n      .rx-virtual-scroll__run-way {\n        width: 1px;\n        height: 1px;\n        contain: strict;\n        position: absolute;\n        will-change: transform;\n      }\n    "], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewportComponent, decorators: [{
            type: Component,
            args: [{
                    selector: 'rx-virtual-scroll-viewport',
                    template: `
    <div #scrollViewport class="rx-virtual-scroll__viewport">
      <div #runway class="rx-virtual-scroll__run-way"></div>
      <ng-content></ng-content>
    </div>
  `,
                    providers: [
                        {
                            provide: RxVirtualScrollViewport,
                            useExisting: RxVirtualScrollViewportComponent,
                        },
                    ],
                    styles: [
                        `
      :host {
        display: block;
        width: 100%;
        height: 100%;
        box-sizing: border-box;
        contain: strict;
      }

      :host:not([autosize]) .rx-virtual-scroll__viewport {
        transform: translateZ(0);
        will-change: scroll-position;
      }

      .rx-virtual-scroll__viewport {
        contain: strict;
        -webkit-overflow-scrolling: touch;
        width: 100%;
        position: absolute;
        top: 0;
        bottom: 0;
        overflow: auto;
      }

      .rx-virtual-scroll__run-way {
        width: 1px;
        height: 1px;
        contain: strict;
        position: absolute;
        will-change: transform;
      }
    `,
                    ],
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }, { type: RxVirtualScrollStrategy, decorators: [{
                    type: Optional
                }] }]; }, propDecorators: { runway: [{
                type: ViewChild,
                args: ['runway', { static: true }]
            }], scrollViewport: [{
                type: ViewChild,
                args: ['scrollViewport', { static: true }]
            }], viewRepeater: [{
                type: ContentChild,
                args: [RxVirtualViewRepeater]
            }], viewRange: [{
                type: Output
            }], scrolledIndexChange: [{
                type: Output
            }] } });

class RxVirtualScrollingModule {
}
/** @nocollapse */ RxVirtualScrollingModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollingModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ RxVirtualScrollingModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollingModule, declarations: [RxVirtualScrollViewportComponent, RxVirtualFor], exports: [RxVirtualScrollViewportComponent, RxVirtualFor] });
/** @nocollapse */ RxVirtualScrollingModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollingModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollingModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [RxVirtualScrollViewportComponent, RxVirtualFor],
                    declarations: [RxVirtualScrollViewportComponent, RxVirtualFor],
                    providers: [],
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { AutosizeVirtualScrollStrategy, AutosizeVirtualScrollStrategyModule, DynamicSizeVirtualScrollStrategy, DynamicSizeVirtualScrollStrategyModule, FixedSizeVirtualScrollStrategy, FixedSizeVirtualScrollStrategyModule, RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS_FACTORY, RxVirtualFor, RxVirtualForViewContext, RxVirtualScrollStrategy, RxVirtualScrollViewportComponent, RxVirtualScrollingModule, RxVirtualViewRepeater };
//# sourceMappingURL=template-experimental-virtual-scrolling.js.map
